#include <stdio.h>
#include <stdio.h>
#define SEED 13516093

void rng(int* arr, int n);
void radixSortSerial(int* arr, int n);
void radixSortParallel(int* arr, int n);
int isMatch(int* arr1, int* arr2, int n);

int main(int argc, char** argv) {
    int n;
    int *arr;
    int *arr_d;
    cudaEvent_t startTime, stopTime;
    float elapsedTime;

    if (argc < 2) {
        printf("Usage: ./radix_sort_cuda <number of array elements>");
        return 0;
    }

    n = strtol(argv[1], NULL, 10);
    arr = (int*)malloc(n*sizeof(int));
    rng(arr, n);
    cudaEventCreate(&startTime);
    cudaEventCreate(&stopTime);


    // Serial Radix Sort
    cudaEventRecord(startTime, 0);

    radixSortSerial(arr, n);

    cudaEventRecord(stopTime, 0);
    cudaEventSynchronize(stopTime);
    cudaEventElapsedTime(&elapsedTime, startTime, stopTime);
    printf("Serial Radix Sort: %lf\n", elapsedTime);


    // CUDA Radix Sort
    cudaMalloc((void**)&arr_d, sizeof(int) * n);
    cudaMemcpy(arr_d, arr, sizeof(int)*n, cudaMemcpyHostToDevice);

    cudaEventRecord(startTime, 0);

    radixSortParallel(arr, n);

    cudaEventRecord(stopTime, 0);
    cudaEventSynchronize(stopTime);
    cudaEventElapsedTime(&elapsedTime, startTime, stopTime);
    printf("Parallel Radix Sort: %lf\n", elapsedTime);

    return 0;
}

void rng(int* arr, int n) {
    srand(SEED);

    for (int i = 0; i < n; i++)
        arr[i] = (int)rand();
}

int getMax(int* arr, int n) {
    int temp = 0;

    for (int i = 0; i < n; i++)
        if (arr[i] > temp) temp = arr[i];

    return temp;
}

// Serial Sort
void countSortSerial(int *arr, int n, int exp) {
    int count[10] = {0};
    int output[n];

    for (int i = 0; i < n; i++) {
        int digit = (arr[i]/exp)%10;
        count[digit]++;
    }

    for (int i = 1; i < 10; i++)
        count[i] += count[i-1];

    for (int i = n-1; i>= 0; i--) {
        int digit = (arr[i]/exp)%10;
        output[count[digit]-1] = arr[i];
        count[digit]--;
    }

    for (int i = 0; i < n; i++)
        arr[i] = output[i];
}

void radixSortSerial(int* arr, int n) {
    int max = getMax(arr, n);

    for (int exp = 1; max/exp > 0; exp *= 10)
        countSortSerial(arr, n, exp);
}

// Parallel Sort
__global__ void countSortParallel(int *arr, int n, int digit) {

}

void radixSortParallel(int *arr, int n) {
    int max = getMax(arr, n);
    int *arr_d;

    cudaMalloc((void**)&arr_d, n*sizeof(int);
    cudaMemcpy(arr_d, arr, n*sizeof(int), cudaMemcpyHostToDevice);

    for (int digit = 1; max/digit > 0; digit *= 10)
        countSort<<<1,1>>>(arr_d, n, digit);

    cudaMemcpy(arr, arr_d, n*sizeof(int), cudaMemcpyDeviceToHost);
}
